package checkers.interning;

import checkers.basetype.*;
import checkers.interning.quals.Interned;
import checkers.interning.quals.PolyInterned;
import checkers.quals.TypeQualifiers;
import checkers.source.*;

/**
 * A typechecker plug-in for the {@link checkers.interning.quals.Interned} qualifier that
 * finds (and verifies the absence of) equality-testing and interning errors.
 *
 * <p>
 *
 * The {@link checkers.interning.quals.Interned} annotation indicates that a variable
 * refers to the canonical instance of an object, meaning that it is safe to
 * compare that object using the "==" operator. This plugin warns whenever
 * "==" is used in cases where one or both operands are not
 * {@link checkers.interning.quals.Interned}.  Optionally, it suggests using "=="
 * instead of ".equals" where possible.
 *
 * @manual #interning-checker Interning checker
 */
@TypeQualifiers({ Interned.class, PolyInterned.class })
@SupportedLintOptions({"dotequals", "flow"})
public final class InterningChecker extends BaseTypeChecker { }
