package checkers.util;

import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.Name;

import checkers.types.QualifierHierarchy;

//
// This is still work in progress class
// It's functional, but requires optimization and better documentation
// I'll come back to this class later
//
/**
 * Represents the type qualifier hierarchy of a type system.
 *
 */
public class DAGQualifierHierarchy extends QualifierHierarchy {

    /**
     * Factory used to create an instance of {@link DAGQualifierHierarchy}.
     */
    public static class Factory {
        /** map: qualifier --> supertypes of the qualifier */
        private Map<AnnotationMirror, Set<AnnotationMirror>> subtypes;

        /** the root of all the qualifiers **/
        private AnnotationMirror root;

        public Factory(DAGQualifierHierarchy tree) {
            subtypes = copy(tree.subtypes);
            root = findRoot(subtypes);
        }

        public Factory() {
            subtypes = AnnotationUtils.createAnnotationMap();
        }

        public void addQualifier(AnnotationMirror qual) {
            if (subtypes.containsKey(qual))
                return;
            subtypes.put(qual, AnnotationUtils.createAnnotationSet());
        }

        public void addSubtype(AnnotationMirror sub, AnnotationMirror sup) {
            addQualifier(sub);
            addQualifier(sup);
            subtypes.get(sub).add(sup);
        }

        public void setRoot(AnnotationMirror root) {
            addQualifier(root);
            this.root = root;
        }

        public DAGQualifierHierarchy build() {
            if (root == null)
                addQualifier(root);
            return new DAGQualifierHierarchy(this);
        }
    }

    /** map: qualifier --> supertypes of the qualifier**/
    private final Map<AnnotationMirror, Set<AnnotationMirror>> subtypes;
    /** the root of all the qualifiers **/
    private final AnnotationMirror root;

    private DAGQualifierHierarchy(Factory f) {
        this.subtypes = copy(f.subtypes);
        this.root = f.root;
//        this.root = findRoot(this.subtypes);
    }

    @Override
    public AnnotationMirror getRootAnnotation() {
        return root;
    }

    @Override
    public Set<Name> getTypeQualifiers() {
        Set<Name> names = new HashSet<Name>();
        for (AnnotationMirror anno: subtypes.keySet())
            names.add(AnnotationUtils.annotationName(anno));
        return Collections.unmodifiableSet(names);
    }

    @Override
    public boolean inConflict(AnnotationMirror anno1, AnnotationMirror anno2) {
        return subtypes.containsKey(anno1) && subtypes.containsKey(anno2);
    }

    @Override
    public AnnotationMirror leastUpperBound(AnnotationMirror a1, AnnotationMirror a2) {
        if (AnnotationUtils.areSame(a1, a2))
            return a1;
        if (AnnotationUtils.areSame(root, a1) || AnnotationUtils.areSame(root, a2))
            return root;
        if (isSubtype(a1, a2))
            return a2;
        else if (isSubtype(a2, a1))
            return a1;
        return root;
    }

    @Override
    public boolean isSubtype(AnnotationMirror anno1, AnnotationMirror anno2) {
        if (AnnotationUtils.areSame(root, anno2))
            return true;
        if (AnnotationUtils.areSameIgnoringValues(anno1, anno2))
            return AnnotationUtils.areSame(anno1, anno2);
        return isSubtypeImp(anno1, anno2);
    }

    private boolean isSubtypeImp(AnnotationMirror anno1, AnnotationMirror anno2) {
        if (AnnotationUtils.areSameIgnoringValues(anno1, anno2))
            return true;
        Set<AnnotationMirror> a1Supers = subtypes.get(anno1);
        if (a1Supers.isEmpty())
            return false;
        for (AnnotationMirror a1Super : a1Supers) {
            if (isSubtypeImp(a1Super, anno2))
                return true;
        }
        return false;
    }

    private static Map<AnnotationMirror, Set<AnnotationMirror>>
    copy(Map<AnnotationMirror, Set<AnnotationMirror>> subtypes) {
        Map<AnnotationMirror, Set<AnnotationMirror>> copy = AnnotationUtils.createAnnotationMap();

        for (AnnotationMirror key : subtypes.keySet()) {
            Set<AnnotationMirror> korig = subtypes.get(key);
            Set<AnnotationMirror> kcopy = AnnotationUtils.createAnnotationSet();
            kcopy.addAll(korig);
            copy.put(key, kcopy);
        }
        return copy;
    }

    private static AnnotationMirror
    findRoot(Map<AnnotationMirror, Set<AnnotationMirror>> subtypes) {
        List<AnnotationMirror> possibleRoots = new LinkedList<AnnotationMirror>();
        for (AnnotationMirror anno : subtypes.keySet()) {
            if (subtypes.get(anno).isEmpty())
                possibleRoots.add(anno);
        }
        assert possibleRoots.size() == 1 : possibleRoots;
        return possibleRoots.get(0);
    }
    private static Map<AnnotationMirror, Set<AnnotationMirror>>
    buildHierarchy(Map<AnnotationMirror, Set<AnnotationMirror>> subtypes) {
        Map<AnnotationMirror, Set<AnnotationMirror>> copy = AnnotationUtils.createAnnotationMap();

        return copy;
    }
}
