import checkers.igj.quals.*;

@I
public class MethodInvocation {

    @Mutable MethodInvocation mutable;
    @Immutable MethodInvocation immutable;
    @ReadOnly MethodInvocation readOnly;

    void mutableMethod() @Mutable { }
    void immutableMethod() @Immutable { }
    void readOnlyMethod() @ReadOnly { }
    void assignsFieldsMethod() @AssignsFields { }

    void testMutable() @Mutable {
        mutableMethod();
        immutableMethod();  // should emit error
        readOnlyMethod();
        assignsFieldsMethod();

        this.mutableMethod();
        this.immutableMethod();  // should emit error
        this.readOnlyMethod();
        this.assignsFieldsMethod();

        mutable.mutableMethod();
        mutable.immutableMethod();  // should emit error
        mutable.readOnlyMethod();
        mutable.assignsFieldsMethod();

        immutable.mutableMethod();  // should emit error
        immutable.immutableMethod();
        immutable.readOnlyMethod();
        immutable.assignsFieldsMethod();  // should emit error

        readOnly.mutableMethod();  // should emit error
        readOnly.immutableMethod();  // should emit error
        readOnly.readOnlyMethod();
        readOnly.assignsFieldsMethod();  // should emit error
    }

    void testImmutable() @Immutable {
        mutableMethod();  // should emit error
        immutableMethod();
        readOnlyMethod();
        assignsFieldsMethod();  // should emit error

        this.mutableMethod();  // should emit error
        this.immutableMethod();
        this.readOnlyMethod();
        this.assignsFieldsMethod();  // should emit error

        mutable.mutableMethod();
        mutable.immutableMethod();  // should emit error
        mutable.readOnlyMethod();
        mutable.assignsFieldsMethod();

        immutable.mutableMethod();  // should emit error
        immutable.immutableMethod();
        immutable.readOnlyMethod();
        immutable.assignsFieldsMethod();  // should emit error

        readOnly.mutableMethod();  // should emit error
        readOnly.immutableMethod();  // should emit error
        readOnly.readOnlyMethod();
        readOnly.assignsFieldsMethod();  // should emit error
    }

    void testReadOnly() @ReadOnly {
        mutableMethod();  // should emit error
        immutableMethod();  // should emit error
        readOnlyMethod();
        assignsFieldsMethod();  // should emit error

        this.mutableMethod();  // should emit error
        this.immutableMethod();  // should emit error
        this.readOnlyMethod();
        this.assignsFieldsMethod();  // should emit error

        mutable.mutableMethod();
        mutable.immutableMethod();  // should emit error
        mutable.readOnlyMethod();
        mutable.assignsFieldsMethod();

        immutable.mutableMethod();  // should emit error
        immutable.immutableMethod();
        immutable.readOnlyMethod();
        immutable.assignsFieldsMethod();  // should emit error

        readOnly.mutableMethod();  // should emit error
        readOnly.immutableMethod();  // should emit error
        readOnly.readOnlyMethod();
        readOnly.assignsFieldsMethod();  // should emit error
    }

    void testAssignsFields() @AssignsFields {
        mutableMethod();  // should emit error
        immutableMethod();  // should emit error
        readOnlyMethod();
        assignsFieldsMethod();

        this.mutableMethod();  // should emit error
        this.immutableMethod();  // should emit error
        this.readOnlyMethod();
        this.assignsFieldsMethod();

        mutable.mutableMethod();
        mutable.immutableMethod();  // should emit error
        mutable.readOnlyMethod();
        mutable.assignsFieldsMethod();

        immutable.mutableMethod();  // should emit error
        immutable.immutableMethod();
        immutable.readOnlyMethod();
        immutable.assignsFieldsMethod();  // should emit error

        readOnly.mutableMethod();  // should emit error
        readOnly.immutableMethod();  // should emit error
        readOnly.readOnlyMethod();
        readOnly.assignsFieldsMethod();  // should emit error
    }

    void testExpressionMutable() {
        (new @Mutable MethodInvocation()).mutableMethod();
        (new @Mutable MethodInvocation()).immutableMethod();  // should emit error
        (new @Mutable MethodInvocation()).readOnlyMethod();
        (new @Mutable MethodInvocation()).assignsFieldsMethod();

        (new @Mutable MethodInvocation()).mutableMethod();
        (new @Mutable MethodInvocation()).immutableMethod();  // should emit error
        (new @Mutable MethodInvocation()).readOnlyMethod();
        (new @Mutable MethodInvocation()).assignsFieldsMethod();

        (true ? mutable : new @Mutable MethodInvocation()).mutableMethod();
        (true ? mutable : new @Mutable MethodInvocation()).immutableMethod();  // should emit error
        (true ? mutable : new @Mutable MethodInvocation()).readOnlyMethod();
        (true ? mutable : new @Mutable MethodInvocation()).assignsFieldsMethod();
    }

    void testExpressionImmutable() {
        (new @Immutable MethodInvocation()).mutableMethod();  // should emit error
        (new @Immutable MethodInvocation()).immutableMethod();
        (new @Immutable MethodInvocation()).readOnlyMethod();
        (new @Immutable MethodInvocation()).assignsFieldsMethod();  // should emit error

        (new @Immutable MethodInvocation()).mutableMethod();  // should emit error
        (new @Immutable MethodInvocation()).immutableMethod();
        (new @Immutable MethodInvocation()).readOnlyMethod();
        (new @Immutable MethodInvocation()).assignsFieldsMethod();  // should emit error

        (true ? immutable : new @Immutable MethodInvocation()).mutableMethod();  // should emit error
        (true ? immutable : new @Immutable MethodInvocation()).immutableMethod();
        (true ? immutable : new @Immutable MethodInvocation()).readOnlyMethod();
        (true ? immutable : new @Immutable MethodInvocation()).assignsFieldsMethod();  // should emit error
    }

    void testExpressionReadOnly() {
        (true ? mutable : new @Immutable MethodInvocation()).mutableMethod();  // should emit error
        (true ? mutable : new @Immutable MethodInvocation()).immutableMethod(); // should emit error
        (true ? mutable : new @Immutable MethodInvocation()).readOnlyMethod();
        (true ? mutable : new @Immutable MethodInvocation()).assignsFieldsMethod();  // should emit error

        (true ? immutable : new @Mutable MethodInvocation()).mutableMethod();  // should emit error
        (true ? immutable : new @Mutable MethodInvocation()).immutableMethod(); // should emit error
        (true ? immutable : new @Mutable MethodInvocation()).readOnlyMethod();
        (true ? immutable : new @Mutable MethodInvocation()).assignsFieldsMethod();  // should emit error
    }
}
